# -*- coding: utf-8 -*-
"""GiocaOnline client module"""
import logging
import time
from copy import deepcopy
from datetime import datetime
from enum import Enum
from hashlib import sha1
from typing import Any, Iterable, List, Mapping, Optional, Union

import pytz
from requests import Response

from thx.giocaonline.client.base_aams_client_mixin import AAMSClientMixin
from thx.giocaonline.constants import GiocaOnlineErrorTypes, PrizePoolType, Rebuy
from thx.giocaonline.exceptions import InternalError, ValidationError

logger = logging.getLogger('giocaonline')

GIOCAONLINE_DATETIME_FORMAT = "%d%m%Y%H%M%S"
GIOCAONLINE_DATE_FORMAT = "%d%m%Y"
EUROBET_DATE_FORMAT = "%Y%m00%d"

PrizeObject = Mapping[str, Union[float, str]]
WinObject = Mapping[str, Union[float, str]]


class GiocaOnlineInternalError(InternalError):
    """Indica un errore di GiocaOnline rispetto alle specifiche indicate in documentazione."""


def strftime(date: datetime):
    return date.strftime(GIOCAONLINE_DATETIME_FORMAT)


def strfdate(date: datetime):
    return date.strftime(GIOCAONLINE_DATE_FORMAT)


def strptime(time_: str):
    return datetime.strptime(time_, GIOCAONLINE_DATETIME_FORMAT)


def strpdate(date: str):
    return datetime.strptime(date, GIOCAONLINE_DATE_FORMAT)


def safe_strpdate(date: str) -> datetime.date:
    """
    Funzione necessaria per gestire i formati data errati inviati dai concessionari

    Errori conosciuti:
        - Sisal: 000000
        - Eurobet: "%Y%m00%d"
    """
    try:
        return strpdate(date)
    except (ValueError, TypeError) as e:
        logger.error(e)
        logger.info("Invalid datetime provided: %s. Trying to parse special cases.", date)

    if date == '000000':  # Sisal fix
        return datetime.now().date()

    try:  # Eurobet fix
        return datetime.strptime(date, EUROBET_DATE_FORMAT)
    except (ValueError, TypeError):
        pass

    return datetime.now().date()


class GiocaOnlineClient(AAMSClientMixin):
    """
    GiocaOnline HTTP client.

    Comunicazioni di gioco
    =========================

    Il modello qui utilizzato è quello che il PGDA deﬁnisce "modalità 2". Le iscrizioni vengono raccolte dopo
     la creazione del torneo (cioè l'apertura sessione).

    Ciascuna dà origine a una richiesta DIRITTO e può venire annullata con una richiesta ANNULLO DIRITTO.
    Al termine delle iscrizioni viene richiesta la CONVALIDA SESSIONE.

    La sequenza delle operazioni è quindi (in maiuscolo l'invocazione dei metodi della presente API):
    • INIZIO SESSIONE
    • inizio iscrizioni
    • DIRITTO / ANNULLO DIRITTO ...
    • ﬁne iscrizioni
    • CONVALIDA SESSIONE
    • PIANO PREMI
    • inizio gioco
    • ﬁne gioco
    • LISTA VINCITORI
    • ACCREDITO ...
    • FINE SESSIONE

    Modalità di trasmissione
    ========================

    Le API verranno invocate dal provider con chiamate HTTPS su un endpoint GOL prestabilito
     (<dominio>/servizi/API.php).
    Si assume che le richieste partano da sistemi identificati (tramite IP o subnet) e abbiano sempre un set di
    parametri obbligatori volti ad autorizzare e identificare il richiedente (definiti separatamente da queste
    specifiche). Richiesta e risposta per ciascun metodo sono codificati come oggetti JSON. Nei nomi delle chiavi,
    le prime due lettere (minuscole) indicano la tipologia di dato, secondo la tabella:

    prefisso   -      descrizione
    -----------------------------------------------------------------------------------------------------------
    fl         -     valore booleano. Si utilizzano le costanti letterali "SI" e "NO", maiuscole
    id        -      identificativo alfanumerico composto esattamente da 16 caratteri
    ip        -      indirizzo IP in notazione dotted-quad
    is        -      identificativo alfanumerico composto esattamente da 14 caratteri
    ls        -      stringa da elenco
    ll        -      importi o valori numerici a 64 bit
    nm        -      valore numerico
    pc        -      valore percentuale, in centesimi (es. 5% => 500)
    sd        -      formato data ridotto. Il valore deve essere formattato come ddmmyyyy (giorno, mese, anno)
    st        -      stringa
    vl        -      valori rappresentanti importi, espressi in centesimi di euro
    xd        -      formato data esteso. Il valore deve essere formattato come ddmmyyyyhhmmss
                    (giorno, mese, anno, ore, minuti, secondi)

    Le comunicazioni inizioSessione, convalidaSessione, pianoPremi, listaVincitori, fineSessione vanno inviate al
    capofila del network.
    In mancanza di una risposta positiva la sessione di gioco non può proseguire.
    Tuttavia una copia delle comunicazioni convalidaSessione e listaVincitori può venire richiesta da altri operatori
     partecipanti al network. L'esito di questo invio, a differenza di quello al capofila, non deve essere bloccante.

    Utilizzare il parametro related_networks nei rispettivi metodi
    """
    current_message: Optional[int] = None
    data: Optional[dict] = None

    STAGING_URL: str = "https://snai.stage.giocaonlinesrl.it/"
    PRODUCTION_URL: str = "https://snai.stage.giocaonlinesrl.it/"  # todo ChangeMe

    _methods_map: dict = {
        # i metodi che iniziano con 1 non sono previsti da AAMS, ma soltanto da GiocaOnline per gestire le
        # comunicazioni fra provider del gioco e concessionario. La numerazione è completamente arbitraria
        # ed interna al pacchetto python.

        100: ("autenticazione_giocatore_100", "autenticazione",),
        101: ("richiesta_saldo_giocatore_101", "statoConto",),
        200: ("inizio_sessione_200", "inizioSessione",),
        220: ("acquisto_diritto_di_partecipazione_220", "dirittoPartecipazione",),
        230: ("richiesta_annullamento_diritto_di_partecipazione_230", "annulloDirittoPartecipazione",),
        240: ("piano_dei_premi_240", "pianoPremi",),
        250: ("convalida_della_sessione_250", "convalidaSessione",),
        260: ("lista_vincitori_260", "listaVincitori",),
        280: ("accredito_vincita_280", "accreditoVincita",),
        300: ("fine_sessione_300", "fineSessione",),
        830: ("software_installato_830", "softwareInstallato",),
    }

    # Request params used by resend request if necessary
    _current_message: int
    method: str
    uri: str
    kwargs: dict

    # Post request params
    related_networks: Optional[List[int]] = None

    def __init__(
        self,
        provider_id: int,
        network_id: int,
        api_key: str,
        requests_params: Optional[dict] = None,
        default_headers: Optional[dict] = None,
        debug: bool = False,
        service_level: str = "TEST",
        resend_request_sleep: float = 0.2,
        base_url: Optional[str] = None
    ) -> None:
        if service_level not in ("TEST", "PROD",):
            raise ValueError("`service_level` must be `TEST` or `PROD`")

        self.check_type('provider_id', provider_id, int)
        self.check_type('network_id', network_id, int)
        self.check_type('api_key', api_key, str)
        self.check_type('resend_request_sleep', resend_request_sleep, float)

        self.provider_id = provider_id
        self.network_id = network_id
        self.api_key = api_key
        self.resend_request_sleep = resend_request_sleep
        if base_url is None:
            base_url = self.STAGING_URL if service_level == "TEST" else self.PRODUCTION_URL
        super().__init__(base_url, requests_params, default_headers, debug)

    def get_default_data(self, id_trans: int):
        """
        La richiesta conterrà sempre almeno il seguente set di chiavi:
            id_fornitore (dealer_id): intero ID del fornitore del gioco (assegnato da GOL)
            id_network: intero ID dell'operatore target (assegnato da GOL). 1 identifica il capofila (Snai)
            id_trans: intero ID univoco della transazione (assegnato dal fornitore)
            method: stringa tipo della richiesta, nella forma aggregation_skill/<richiesta>
            key: stringa API key (indipendente da method)
        """
        self.check_type('id_trans', id_trans, int)
        method = self.get_gol_method(self.current_message)
        return {
            "id_fornitore": self.provider_id,
            "id_network": self.network_id,
            "id_trans": id_trans,
            "method": f"aggregation_skill/{method}",
            "key": self.api_key,
        }

    def get_next_id_transaction(self):
        """Hook this method to provide a custom method to get the next id transaction"""
        return self.id_transaction + 1

    def get_method(self, message_code: int, default=None) -> str:
        return self._methods_map[message_code][0]

    def get_gol_method(self, message_code: int) -> str:
        return self._methods_map[message_code][1]

    @property
    def id_transaction(self):
        return (self.data or {}).get("id_trans", -1)

    def get_endpoint(self, message_code):
        return "servizi/API.php"

    def validate_and_prepare_data(self, message_code, **kwargs):
        id_trans = int(kwargs.pop("id_trans"))
        super().validate_and_prepare_data(message_code, **kwargs)
        self.data.update(self.get_default_data(id_trans))

    def do_request(self, method: str, uri: str, **kwargs) -> None:
        try:
            self.pre_request(method, uri, **kwargs)
            self.response = getattr(self.session, method)(uri, **kwargs)
            self.post_request(method, uri, **kwargs)
        except Exception as e:
            logger.error(e)
            raise e

    def pre_request(self, method: str, uri: str, **kwargs) -> None:
        # Set data for eventually pass by _resend_request() method
        self._current_message = self.current_message
        self.method = method
        self.uri = uri
        self.kwargs = kwargs

    def post_request(self, method: str, uri: str, **kwargs) -> None:
        self.send_related_networks_requests(method, uri, **kwargs)

    def send_related_networks_requests(self, method: str, uri: str, **kwargs):
        if self.related_networks:
            data = deepcopy(kwargs)
            for network_id in self.related_networks:
                try:
                    data = self._prepare_related_network_request_data(network_id, data)
                    getattr(self.session, method)(uri, **data)
                except Exception as e:
                    logger.error(e)

            self.related_networks = None

    def _prepare_related_network_request_data(self, id_network, data):
        data["json"]["id_network"] = id_network
        data["json"]["id_trans"] = self.get_next_id_transaction()
        return data

    # MESSAGGI INTERNI PROVIDER - GIOCAONLINE - CONCESSIONARIO
    def autenticazione_giocatore_100(
        self,
        id_trans: int,
        player_token: str,
        player_account_code: str,
        player_ip: str,
    ):
        """
        Richiesta di verifica del token passato dal concessionario

        :param id_trans: ID univoco della transazione (assegnato dal fornitore)
        :param player_token: (stGiocatore) codice univoco (token) assegnato dal concessionario per il lancio del gioco
        :param player_account_code: (stConto) codice conto del giocatore
        :param player_ip: (ipGiocatore) indirizzo IP del giocatore
        """
        return self._do_request_message(
            message_code=100,
            id_trans=id_trans,
            player_token=player_token,
            player_account_code=player_account_code,
            player_ip=player_ip,
        )

    def validate_and_prepare_data_100(
        self,
        player_token: str,
        player_account_code: str,
        player_ip: str,
    ):
        self.check_type('player_token', player_token, str)
        self.check_player_account_code(player_account_code)
        self.check_type('player_ip', player_ip, str)

        self.data = {
            "stGiocatore": player_token,
            "stConto": player_account_code,
            "ipGiocatore": player_ip,
        }

    def handle_message_100_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 100.

        Parametri in risposta:
        =====================

        {
            "account_code": "abc", # (stConto) codice conto del giocatore
            "player_ssn": "abc", # (idCodFisc) codice fiscale del giocatore (o suo hash)
            "region": "abc", # (lsRegione) codice regione del giocatore
            "reseller_id": "abc", # (nmRivenditore) (opzionale) ID del rivenditore
            "access_token": "abc", # (stAccesso) codice di accesso (token) da impiegare per le successive richieste
                                      sulla piattaforma GOL,
        }

        NB: reseller_id (nmRivenditore) è un codice numerico (intero a 32 bit) utilizzabile dagli aggregatori
        per indicare il concessionario di appartenenza del giocatore.
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                return {
                    "account_code": msg_or_exception["stConto"],
                    "player_ssn": msg_or_exception["idCodFisc"],
                    "region": msg_or_exception["lsRegione"],
                    "access_token": msg_or_exception["stAccesso"],
                    "reseller_id": msg_or_exception.get("nmRivenditore", None),
                }
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)
            except Exception as e:
                self.raise_response_generic_error(e, msg_or_exception)

        raise msg_or_exception

# ================================================================================
    def richiesta_saldo_giocatore_101(
        self,
        id_trans: int,
        gol_player_token: str,
        player_account_code: str,
        play_code: int,
    ):
        """
        Richiesta di saldo del conto giocatore

        :param id_trans: ID univoco della transazione (assegnato dal fornitore)
        :param gol_player_token: (stAccesso) codice di accesso (token) assegnato al giocatore sulla piattaforma GOL
        :param player_account_code: (stConto) codice conto del giocatore
        :param play_code: (nmCodiceGioco) codice gioco AAMS
        """
        return self._do_request_message(
            message_code=101,
            id_trans=id_trans,
            gol_player_token=gol_player_token,
            player_account_code=player_account_code,
            play_code=play_code,
        )

    def validate_and_prepare_data_101(
        self,
        gol_player_token: str,
        player_account_code: str,
        play_code: int,
    ):
        self.check_gol_player_token(gol_player_token)
        self.check_play_code(play_code)
        self.check_player_account_code(player_account_code)

        self.data = {
            "stAccesso": gol_player_token,
            "stConto": player_account_code,
            "nmCodiceGioco": play_code,
        }

    def handle_message_101_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 101.

        Parametri in risposta:
        =====================

        {
            "account_amount": "11.22", # (vlSaldo) saldo del conto
            "account_bonus_amount": "11.22", # (vlSaldoBonus) (opzionale) saldo real bonus
            "account_play_bonus_amount": "11.22", # (vlSaldoPlayBonus) (opzionale) saldo bonus
        }

        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                return {
                    "account_amount": self.convert_cent_value_to_float(msg_or_exception["vlSaldo"]),
                    "account_bonus_amount": self.convert_cent_value_to_float(msg_or_exception.get("vlSaldoBonus", 0)),
                    "account_play_bonus_amount": self.convert_cent_value_to_float(
                        msg_or_exception.get("vlSaldoPlayBonus", 0)
                    ),
                }
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)
            except Exception as e:
                self.raise_response_generic_error(e, msg_or_exception)

        raise msg_or_exception

# ================================================================================
    # MESSAGGI PER LO SVOLGIMENTO DEL GIOCO IN MODALITA’ 2. A partire da pagina 30 della documentazione ufficiale.

    # 200
    def inizio_sessione_200(
        self,
        id_trans: int,
        session_id: str,
        play_code: int,
        session_datetime_start: datetime,
        session_datetime_end: datetime,
        imp_diritto_partecip: float,
        prize_pool_type: PrizePoolType,
        prize_pool_perc: float,  # rtp
        num_expected_players: int = 0,  # valorizzare sempre a 0 per ADK (mai un SIT and GO)
        communication_modality: bool = True,
        minimum_guaranteed_sum: float = 0.0,
        buy_back_tickets_allowed: bool = False,
        is_bonus: bool = False,
        restricted_winnings: bool = False,  # True solo per i tornei SATELLITE
        int_jackpot: bool = False,
        additional_jackpot: bool = False,
        cup_modality: bool = False
    ):
        """
        Creazione di un nuovo torneo.

        :param id_trans: ID univoco della transazione (assegnato dal fornitore)
        :param session_id: (idSessConc) identificativo univoco della sessione sui sistemi del fornitore
        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param session_datetime_start: (xdInizio) data inizio sessione (UTC)
        :param session_datetime_end: (sdFine) data fine presunta sessione (UTC)
        :param communication_modality: (flModCondizionata) Valorizzare sempre a SI
        :param imp_diritto_partecip: (vlImporto) importo diritto di partecipazione
        :param prize_pool_type: (lsTpMtPr) tipologia di montepremi PERCENTUALE, VARIABILE, SOMMA_MIN_GARANTITA
        :param prize_pool_perc: (pcMontepremi) percentuale montepremi (es 5% --> 0.05)
        :param minimum_guaranteed_sum: (vlSommaMinGarantita) somma minima garantita
        :param buy_back_tickets_allowed: (lsRiacqBigl) sessione in cui sono consentiti acquisti multipli di biglietti.
            Passare un valore booleano che verrà trasformato in (NESSUNO (False) o REBUY (True))
        :param num_expected_players: (nmGiocatoriPrevisti) numero giocatori previsti
        :param is_bonus: (flBonus) sessione con bonus. Valorizzare a SI.
        :param restricted_winnings: (flSessVincColl) sessione con vincite vincolate
        :param int_jackpot: (flJckInterno) sessione con jackpot interno al gioco
        :param additional_jackpot: (flJckAgg)  sessione con jackpot aggiuntivi
        :param cup_modality: (flCup) sessione di gioco in modalità CUP. Valorizzare sempre a NO.

        Note
        ===========
        • Se lsTpMtPr=VARIABILE, pcMontepremi indica la percentuale media del montepremi (> 8000).
        • vlSommaMinGarantita è obbligatorio se lsTpMtPr=SOMMA_MIN_GARANTITA.
            # TODO posso passare di default 0 nei casi di torneo non garantito?
        • I valori consentiti per lsRiacqBigl sono NESSUNO (False) o REBUY (True).
        • nmGiocatoriPrevisti deve essere valorizzato al numero di giocatori previsti se il torneo è sit and go,
            altrimenti a 0.
        • flSessVincColl deve essere valorizzato a SI per i tornei satellite.

        """
        return self._do_request_message(
            message_code=200,
            id_trans=id_trans,
            session_id=session_id,
            play_code=play_code,
            session_datetime_start=session_datetime_start,
            session_datetime_end=session_datetime_end,
            communication_modality=communication_modality,
            imp_diritto_partecip=imp_diritto_partecip,
            prize_pool_type=prize_pool_type,
            prize_pool_perc=prize_pool_perc,
            minimum_guaranteed_sum=minimum_guaranteed_sum,
            buy_back_tickets_allowed=buy_back_tickets_allowed,
            num_expected_players=num_expected_players,
            is_bonus=is_bonus,
            restricted_winnings=restricted_winnings,
            int_jackpot=int_jackpot,
            additional_jackpot=additional_jackpot,
            cup_modality=cup_modality,
        )

    def validate_and_prepare_data_200(
        self,
        session_id: str,
        play_code: int,
        session_datetime_start: datetime,
        session_datetime_end: datetime,
        imp_diritto_partecip: float,
        prize_pool_type: PrizePoolType,
        prize_pool_perc: float,  # rtp
        num_expected_players: int,
        communication_modality: bool = True,
        minimum_guaranteed_sum: float = 0.0,
        buy_back_tickets_allowed: bool = False,
        is_bonus: bool = False,
        restricted_winnings: bool = False,
        int_jackpot: bool = False,
        additional_jackpot: bool = False,
        cup_modality: bool = False
    ):
        self.check_aams_id(session_id)
        self.check_play_code(play_code)
        self.check_utc_timezone("session_datetime_start (xdInizio)", session_datetime_start)
        self.check_utc_timezone("session_datetime_end (xdInizio)", session_datetime_end)
        self.check_positive_amount("imp_diritto_partecip (vlImporto)", imp_diritto_partecip)
        # verifica che la tipologia di montepremi sia una delle opzioni valide
        self.check_enum("prize_pool_type (lsTpMtPr)", prize_pool_type, PrizePoolType)
        self.check_perc_amount("prize_pool_perc", prize_pool_perc)
        self.check_type("num_expected_players", num_expected_players, int)
        # verifica che il numero di giocatori previsti sia superiore o uguale a 0
        self.check_numeric_value("num_expected_players (nmGiocatoriPrevisti)", num_expected_players, 0, "ge")
        # verifica che il parametro communication_modality sia esattamente uguale a "SI"
        self.check_value_in("communication_modality (flModCondizionata)", communication_modality, [True])
        # verifica che, se il tipo di torneo è a somma minima garantita, allora il valore di
        # `minimum_guaranteed_sum` (vlSommaMinGarantita) deve essere maggiore di zero, altrimenti 0
        if self.get_enum_value(prize_pool_type) == PrizePoolType.MINIMUM_GUARANTEED_SUM.value:
            self.check_numeric_value("minimum_guaranteed_sum (vlSommaMinGarantita)", minimum_guaranteed_sum, 0, "g")
        else:
            self.check_numeric_value("minimum_guaranteed_sum (vlSommaMinGarantita)", minimum_guaranteed_sum, 0, "e")
        # verifica che, `minimum_guaranteed_sum` abbia al massimo 2 decimali
        self.check_decimal_digits("minimum_guaranteed_sum (vlSommaMinGarantita)", minimum_guaranteed_sum, 2, "le")
        self.check_type("buy_back_tickets_allowed", buy_back_tickets_allowed, bool)
        self.check_type("is_bonus", is_bonus, bool)
        self.check_type("restricted_winnings", restricted_winnings, bool)
        self.check_type("int_jackpot", int_jackpot, bool)
        self.check_type("additional_jackpot", additional_jackpot, bool)
        self.check_type("cup_modality (flModCondizionata)", cup_modality, bool)
        # verifica che il parametro cup_modality sia esattamente uguale a "NO"
        self.check_value_in("cup_modality (flModCondizionata)", cup_modality, [False])

        self.data = {
            "idSessConc": session_id,
            "nmCodiceGioco": play_code,
            "xdInizio": strftime(session_datetime_start),
            "sdFine": strfdate(session_datetime_end),
            "flModCondizionata": self.convert_bool_to_fl(communication_modality),
            "vlImporto": self.convert_float_value_to_cent(imp_diritto_partecip),
            "lsTpMtPr": self.get_enum_value(prize_pool_type),
            "pcMontepremi": self.convert_float_perc_to_cent(prize_pool_perc),
            "lsRiacqBigl": self.get_enum_value(Rebuy.REBUY if buy_back_tickets_allowed else Rebuy.NONE),
            "nmGiocatoriPrevisti": num_expected_players,
            "flBonus": self.convert_bool_to_fl(is_bonus),
            "flSessVincColl": self.convert_bool_to_fl(restricted_winnings),
            "flJckInterno": self.convert_bool_to_fl(int_jackpot),
            "flJckAgg": self.convert_bool_to_fl(additional_jackpot),
            "flCup": self.convert_bool_to_fl(cup_modality),
        }
        if self.get_enum_value(prize_pool_type) == PrizePoolType.MINIMUM_GUARANTEED_SUM.value:
            self.data["vlSommaMinGarantita"] = self.convert_float_value_to_cent(minimum_guaranteed_sum)

    def handle_message_200_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 200.

        Parametri in risposta:
        =====================

        session_id (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                return {"session_id": msg_or_exception["idSessAams"]}
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)
            except Exception as e:
                self.raise_response_generic_error(e, msg_or_exception)
        raise msg_or_exception

# =============================================================================================================
    # 220
    def acquisto_diritto_di_partecipazione_220(
        self,
        id_trans: int,
        play_code: int,
        gol_player_token: str,
        player_account_code: str,
        session_id: str,
        region: str,
        price: float,
        rebuy: str,
        player_ip: str,
        player_nickname: str,
        purchase_datetime: datetime,
        amount_for_jackpot: float,
        rake: float,
        tournament_id: int,
        device: int,
        related_session_id: str = " " * 16,
        pay_with_token: bool = False,
    ):
        """
        Richiesta di acquisto di una schedina.

        :param id_trans: ID univoco della transazione (assegnato dal fornitore)
        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param gol_player_token: (stAccesso) codice di accesso (token) assegnato al giocatore sulla piattaforma GOL
        :param player_account_code: (stConto) codice conto del giocatore
        :param session_id: (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        :param region: (lsRegione) codice regione del giocatore
        :param price: (vlImporto) importo diritto partecipazione
        :param rebuy: (lsRiacqBigl) riacquisto
        :param player_ip: (ipGiocatore) indirizzo IP del giocatore
        :param player_nickname: (stAlias) alias giocatore
        :param purchase_datetime: (xdAcquisto) data acquisto diritto (UTC)
        :param related_session_id: (idSessCollegata) codice identificativo AAMS della sessione collegata (torneo
            satellite vinto)
        :param amount_for_jackpot: (vlImpoJck) importo destinato al fondo jackpot
        :param pay_with_token: (flToken) acquisto con token
        :param rake: (vlRake) importo rake
        :param tournament_id: (stTorneo) codice univoco del torneo sui sistemi del provider
        :param device: (nmTipoDisp) tipo dispositivo

        Note
        ===========
            • idSessCollegata in caso di iscrizione ad un torneo principale dopo la vittoria in un torneo satellite va
              posto uguale al codice sessione AAMS di quest'ultimo. In caso contrario va impostato al valore di 16
              spazi.
            • lsRiacqBigl può valere NESSUNO o, per i tornei che consentono il riacquisto, REBUY
            • nmTipoDisp contiene la descrizione del dispositivo utilizzato dall'utente per l'accesso al gioco, secondo
              la codifica dettagliata nella documentazione PGDA ("Tabella dei tipi di terminali di accettazione gioco")
        """
        return self._do_request_message(
            message_code=220,
            id_trans=id_trans,
            play_code=play_code,
            gol_player_token=gol_player_token,
            player_account_code=player_account_code,
            session_id=session_id,
            region=region,
            price=price,
            rebuy=rebuy,
            player_ip=player_ip,
            player_nickname=player_nickname,
            purchase_datetime=purchase_datetime,
            related_session_id=related_session_id,
            amount_for_jackpot=amount_for_jackpot,
            rake=rake,
            tournament_id=tournament_id,
            device=device,
            pay_with_token=pay_with_token
        )

    def validate_and_prepare_data_220(
        self,
        play_code: int,
        gol_player_token: str,
        player_account_code: str,
        session_id: str,
        region: str,
        price: float,
        rebuy: str,
        player_ip: str,
        player_nickname: str,
        purchase_datetime: datetime,
        related_session_id: str,
        amount_for_jackpot: float,
        rake: float,
        tournament_id: int,
        device: int,
        pay_with_token: bool = False,
    ):
        self.check_play_code(play_code)
        self.check_gol_player_token(gol_player_token)
        self.check_player_account_code(player_account_code)
        self.check_aams_id(session_id)
        self.check_type("region", region, str)
        self.check_positive_amount("price", price)
        self.check_enum("rebuy", rebuy, Rebuy)
        self.check_type("player_ip", player_ip, str)
        self.check_type("player_nickname", player_nickname, str)
        self.check_utc_timezone("purchase_datetime (xdAcquisto)", purchase_datetime,)
        self.check_aams_id(related_session_id, 'related_session_id (idSessCollegata)')
        self.check_positive_amount("amount_for_jackpot", amount_for_jackpot)
        self.check_positive_amount("rake", rake)  # TODO VERIFICARE
        self.check_type("device", device, int)

        self.data = {
            "nmCodiceGioco": play_code,
            "stAccesso": gol_player_token,
            "stConto": player_account_code,
            "idSessAams": session_id,
            "lsRegione": region,
            "vlImporto": self.convert_float_value_to_cent(price),
            "lsRiacqBigl": self.get_enum_value(rebuy),
            "ipGiocatore": player_ip,
            "stAlias": player_nickname,
            "xdAcquisto": strftime(purchase_datetime),
            "idSessCollegata": related_session_id,
            "vlImpoJck": self.convert_float_value_to_cent(amount_for_jackpot),
            "vlRake": self.convert_float_value_to_cent(rake),
            "stTorneo": str(tournament_id),
            "nmTipoDisp": device,
            "flToken": self.convert_bool_to_fl(pay_with_token),
        }

    def handle_message_220_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 220.

        Parametri in risposta:
        =====================

        played_aams_id (idBiglietto) codice univoco del diritto di partecipazione assegnato dal sistema AAMS
        account_amount (vlSaldoCarta) saldo conto di gioco
        account_bonus_amount (vlSaldoBonus) (opzionale) saldo real bonus
        received_by_aams_datetime (sdTimeStamp) data di acquisizione da parte di AAMS
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                response_dict = {
                    "played_aams_id": msg_or_exception["idBiglietto"],
                    "account_amount": self.convert_cent_value_to_float(msg_or_exception["vlSaldoCarta"]),
                    "account_bonus_amount": self.convert_cent_value_to_float(msg_or_exception.get("vlSaldoBonus", 0)),
                    "received_by_aams_datetime": safe_strpdate(msg_or_exception["sdTimeStamp"]),
                }
                return response_dict
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)
            except Exception as e:
                self.raise_response_generic_error(e, msg_or_exception)
        raise msg_or_exception

# =============================================================================================================
    # 230
    def richiesta_annullamento_diritto_di_partecipazione_230(
        self,
        id_trans: int,
        play_code: int,
        gol_player_token: str,
        player_account_code: str,
        session_id: str,
        played_aams_id: str,
        pay_with_token: bool = False,
    ):
        """
        Richiesta di annullamento di un biglietto acquistato in precedenza.

        :param id_trans: ID univoco della transazione (assegnato dal fornitore)
        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param gol_player_token: (stAccesso) codice di accesso (token) assegnato al giocatore sulla piattaforma GOL
        :param player_account_code: (stConto) codice conto del giocatore
        :param session_id: (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        :param played_aams_id: (idBiglietto) identificativo diritto di partecipazione assegnato dal sistema AAMS
        :param pay_with_token: (flToken) acquisto con token

        """
        return self._do_request_message(
            message_code=230,
            id_trans=id_trans,
            play_code=play_code,
            gol_player_token=gol_player_token,
            player_account_code=player_account_code,
            session_id=session_id,
            played_aams_id=played_aams_id,
            pay_with_token=pay_with_token
        )

    def validate_and_prepare_data_230(
        self,
        play_code: int,
        gol_player_token: str,
        player_account_code: str,
        session_id: str,
        played_aams_id: str,
        pay_with_token: bool = False,
    ):
        self.check_play_code(play_code)
        self.check_gol_player_token(gol_player_token)
        self.check_player_account_code(player_account_code)
        self.check_aams_id(session_id)
        self.check_aams_id(played_aams_id, "played_aams_id (idBiglietto)")
        self.check_type("pay_with_token", pay_with_token, bool)

        self.data = {
            "nmCodiceGioco": play_code,
            "stAccesso": gol_player_token,
            "stConto": player_account_code,
            "idSessAams": session_id,
            "idBiglietto": played_aams_id,
            "flToken": self.convert_bool_to_fl(pay_with_token),
        }

    def handle_message_230_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 230.

        Parametri in risposta:
        =====================

        account_amount (vlSaldoCarta) saldo conto di gioco
        account_bonus_amount (vlSaldoBonus) (opzionale) saldo real bonus
        received_by_aams_datetime (sdTimeStamp) data di acquisizione da parte di AAMS
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                response_dict = {
                    "account_amount": self.convert_cent_value_to_float(msg_or_exception["vlSaldoCarta"]),
                    "account_bonus_amount": self.convert_cent_value_to_float(msg_or_exception.get("vlSaldoBonus", 0)),
                    "received_by_aams_datetime": safe_strpdate(msg_or_exception["sdTimeStamp"]),
                }
                return response_dict
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)
            except Exception as e:
                self.raise_response_generic_error(e, msg_or_exception)
        raise msg_or_exception

# =============================================================================================================
    # 240
    def piano_dei_premi_240(
        self,
        id_trans: int,
        play_code: int,
        session_id: str,
        prize_pool_perc: float,
        prize_pool_tot: float,
        internal_jackpot_tot: float,
        total_add_jackpot: float,
        total_prize_number: int,
        initial_prize: int,
        final_prize: int,
        prize_pool_part: float,
        internal_jackpot_part: float,
        partial_add_jackpot: float,
        rectification_progressive_number: int,
        prizes_list: Iterable[PrizeObject],
    ):
        """
        Definizione dei premi da distribuire al termine del gioco

        :param id_trans: ID univoco della transazione (assegnato dal fornitore)
        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param session_id: (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        :param prize_pool_perc: (pcMontepremi) percentuale montepremi erogato
        :param prize_pool_tot: (vlMontepremiTot) importo totale del montepremi erogato, comprensivo di eventuali jackpot
        :param internal_jackpot_tot: (vlImpoJckInterno) quota del montepremi totale derivante da jackpot interno al gioco
        :param total_add_jackpot: (vlImpoJckAgg) quota del montepremi totale derivante da jackpot aggiuntivi
        :param total_prize_number: (nmTotPremi) numero totale premi da attribuire nella sessione di gioco
        :param initial_prize: (nmPremioIniz) primo premio relativo all'invio corrente
        :param final_prize: (nmPremioFinale) ultimo premio relativo all'invio corrente
        :param prize_pool_part: (vlMontepremiParz) importo montepremi relativo ai premi presenti nell'invio
        :param internal_jackpot_part: (vlJckInternoParz) quota del montepremi parziale derivante da jackpot interno al
            gioco
        :param partial_add_jackpot: (vlJckAggParz) quota del montepremi parziale derivante da jackpot aggiuntivi
        :param rectification_progressive_number: (nmProgrRettifica) 0 per il primo invio del piano premi, va incrementato
            di 1 per ogni rettifica
        :param prizes_list: (premi) lista di oggetti premio (ciascuno rappresenta un premio)

        Struttura oggetto premio
        ========================
        vlImpoMontepremi (vlImpoMontepremi) valore del premio, comprensivo di eventuali jackpot
        vlImpoJckInterno (vlImpoJckInterno) quota del premio derivante da jackpot interno al gioco
        vlImpoJckAgg (vlImpoJckAgg) quota del premio derivante da jackpot aggiuntivi

        Note
        ===========
            • I premi presenti nella lista premi vanno indicati in ordine di importo decrescente.
            • Nel caso i premi siano più di 1000, sarà necessario spezzare la lista su più invii, ciascuno con al più
                1000 premi.

        """
        return self._do_request_message(
            message_code=240,
            id_trans=id_trans,
            play_code=play_code,
            session_id=session_id,
            prize_pool_perc=prize_pool_perc,
            prize_pool_tot=prize_pool_tot,
            internal_jackpot_tot=internal_jackpot_tot,
            total_add_jackpot=total_add_jackpot,
            total_prize_number=total_prize_number,
            initial_prize=initial_prize,
            final_prize=final_prize,
            prize_pool_part=prize_pool_part,
            internal_jackpot_part=internal_jackpot_part,
            partial_add_jackpot=partial_add_jackpot,
            rectification_progressive_number=rectification_progressive_number,
            prizes_list=prizes_list,
        )

    def validate_and_prepare_data_240(
        self,
        play_code: int,
        session_id: str,
        prize_pool_perc: float,
        prize_pool_tot: float,
        internal_jackpot_tot: float,
        total_add_jackpot: float,
        total_prize_number: int,
        initial_prize: int,
        final_prize: int,
        prize_pool_part: float,
        internal_jackpot_part: float,
        partial_add_jackpot: float,
        rectification_progressive_number: int,
        prizes_list: List[PrizeObject],
    ):
        self.check_play_code(play_code)
        self.check_aams_id(session_id)
        self.check_perc_amount("prize_pool_perc", prize_pool_perc)
        self.check_positive_amount("prize_pool_tot", prize_pool_tot)
        self.check_positive_amount("internal_jackpot_tot", internal_jackpot_tot)
        self.check_positive_amount("total_add_jackpot", total_add_jackpot)
        self.check_type("total_prize_number", total_prize_number, int)
        self.check_type("initial_prize", initial_prize, int)
        self.check_type("final_prize", final_prize, int)
        self.check_positive_amount("prize_pool_part", prize_pool_part)
        self.check_positive_amount("internal_jackpot_part", internal_jackpot_part)
        self.check_positive_amount("partial_add_jackpot", partial_add_jackpot)
        self.check_type("rectification_progressive_number", rectification_progressive_number, int)

        # Controlla che la lista premi non contenga più di 1000 elementi
        self.check_prizes_list("prizes_list (premi)", prizes_list, 1000)

        self.data = {
            "nmCodiceGioco": play_code,
            "idSessAams": session_id,
            "pcMontepremi": self.convert_float_perc_to_cent(prize_pool_perc),
            "vlMontepremiTot": self.convert_float_value_to_cent(prize_pool_tot),
            "vlImpoJckInterno": self.convert_float_value_to_cent(internal_jackpot_tot),
            "vlImpoJckAgg": self.convert_float_value_to_cent(total_add_jackpot),
            "nmTotPremi": total_prize_number,
            "nmPremioIniz": initial_prize,
            "nmPremioFinale": final_prize,
            "vlMontepremiParz": self.convert_float_value_to_cent(prize_pool_part),
            "vlJckInternoParz": self.convert_float_value_to_cent(internal_jackpot_part),
            "vlJckAggParz": self.convert_float_value_to_cent(partial_add_jackpot),
            "nmProgrRettifica": rectification_progressive_number,
            "premi": self.prepare_prizes_list(prizes_list),
        }

    def prepare_prizes_list(self, prizes_list: List[WinObject]) -> List:
        # Ordina la lista premi in ordine decrescente rispetto all'importo totale della vincita
        prizes_list.sort(key=lambda x: x.get("amount"), reverse=True)
        try:
            result = []
            for prize in prizes_list:
                total_prize = prize["amount"]
                internal_jackpot = prize.get("internal_jackpot", 0)
                additional_jackpot = prize.get("additional_jackpot", 0)

                self.check_positive_amount("total_prize", total_prize)
                self.check_positive_amount("internal_jackpot", internal_jackpot)
                self.check_positive_amount("additional_jackpot", additional_jackpot)

                result.append({
                    "vlImpoMontepremi": self.convert_float_value_to_cent(total_prize),
                    "vlImpoJckInterno": self.convert_float_value_to_cent(internal_jackpot),
                    "vlImpoJckAgg": self.convert_float_value_to_cent(additional_jackpot),
                })
            return result
        except KeyError as e:
            raise ValidationError({str(e): "This key is required"}) from e
        except Exception as e:
            raise ValidationError(str(e)) from e

    def handle_message_240_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 240.

        Parametri in risposta:
        =====================

        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                return {}
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)
            except Exception as e:
                self.raise_response_generic_error(e, msg_or_exception)
        raise msg_or_exception

# =============================================================================================================
    # 250
    def convalida_della_sessione_250(
        self,
        id_trans: int,
        play_code: int,
        session_id: str,
        validation_date: datetime,
        related_networks: Optional[List[int]] = None
    ):
        """
        Inviato alla partenza effettiva del torneo.

        :param id_trans: ID univoco della transazione (assegnato dal fornitore)
        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param session_id: (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        :param validation_date: (xdConvalida) timestamp della convalida sessione (UTC)
        :param related_networks: Lista di ID concessionari a cui inviare la medesima comunicazione
        """
        self.related_networks = related_networks
        self.check_related_networks(related_networks)
        return self._do_request_message(
            message_code=250,
            id_trans=id_trans,
            play_code=play_code,
            session_id=session_id,
            validation_date=validation_date,
        )

    def validate_and_prepare_data_250(
        self,
        play_code: int,
        session_id: str,
        validation_date: datetime,
    ):
        self.check_play_code(play_code)
        self.check_aams_id(session_id)
        self.check_utc_timezone("validation_date (xdConvalida)", validation_date)

        self.data = {
            "nmCodiceGioco": play_code,
            "idSessAams": session_id,
            "xdConvalida": strftime(validation_date),
        }

    def handle_message_250_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 250.

        Parametri in risposta:
        =====================

        received_by_aams_datetime (sdTimeStamp) data di acquisizione da parte di AAMS
        number_of_participation_validated (nmPartConvalidati) numero di partecipazioni convalidate da AAMS
        total_amount_of_participation_validated (vlImpPartConvalidati) importo totale raccolto con le partecipazioni
            convalidate
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                response_dict = {
                    "received_by_aams_datetime": safe_strpdate(msg_or_exception["sdTimeStamp"]),
                    "number_of_participation_validated": msg_or_exception["nmPartConvalidati"],
                    "total_amount_of_participation_validated": self.convert_cent_value_to_float(
                        msg_or_exception["vlImpPartConvalidati"]
                    ),
                }
                return response_dict
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)
            except Exception as e:
                self.raise_response_generic_error(e, msg_or_exception)
        raise msg_or_exception

# =============================================================================================================
    # 260
    def lista_vincitori_260(
        self,
        id_trans: int,
        play_code: int,
        session_id: str,
        total_winner_number: int,
        initial_winner: int,
        final_winner: int,
        win_prize_part: float,
        internal_jackpot_part: float,
        additional_jackpot_part: float,
        rectification_progressive_number: int,
        total_win_prize: float,
        total_internal_jackpot: float,
        total_additional_jackpot: float,
        win_list: List[WinObject],
        related_networks: Optional[List[int]] = None
    ):
        """
        Comunicazione delle vincite realizzate in una sessione di gioco

        :param id_trans: ID univoco della transazione (assegnato dal fornitore)
        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param session_id: (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        :param total_winner_number (nmTotVincitori) numero totale vincitori nella sessione di gioco
        :param initial_winner (nmInizVincita) progressivo prima vincita relativa al presente invio
        :param final_winner (nmFinVincita) progressivo ultima vincita relativa al presente invio
        :param win_prize_part (vlImpoParzVincite) importo relativo alle vincite, comprensivo di jackpot e bonus
        :param internal_jackpot_part (vlImpoParzJckInterno) quota delle vincite derivanti da jackpot interno al gioco
        :param additional_jackpot_part (vlImpoParzJckAgg) quota delle vincite derivanti da jackpot aggiuntivo
        :param rectification_progressive_number (nmProgrRettifica) 0 per il primo invio, va incrementato di 1 per ogni
            richiesta di rettifica
        :param total_win_prize (vlImpoTotVincite) importo totale delle vincite erogate comprensivo di jackpot e bonus
        :param total_internal_jackpot (vlImpoTotJckInterno) quota delle vincite erogate derivante da jackpot interno
            al gioco
        :param total_additional_jackpot (vlImpoTotJckAgg) quota delle vincite erogate derivante da jackpot aggiuntivo
        :param win_list (vincite) lista di oggetti vincita
        :param related_networks: Lista di ID concessionari a cui inviare la medesima comunicazione

        Struttura oggetto vincita
        =========================
        total_prize (vlImpoTotale) importo totale della vincita compresi jackpot e bonus
        internal_jackpot (vlImpoJckInterno) quota della vincita derivante da jackpot interno al gioco
        additional_jackpot (vlImpoJckAgg) quota della vincita derivante da jackpot aggiuntivo
        played_aams_id (idBiglietto) codice biglietto

        Note
        ========
            • Le vincite presenti nella lista vincite vanno indicate in ordine di importo totale decrescente.
            • Nel caso le vincite siano più di 1000, sarà necessario spezzare la lista su più invii, ciascuno con al più
                1000 vincite.
        """
        self.related_networks = related_networks
        self.check_related_networks(related_networks)
        return self._do_request_message(
            message_code=260,
            id_trans=id_trans,
            play_code=play_code,
            session_id=session_id,
            total_winner_number=total_winner_number,
            initial_winner=initial_winner,
            final_winner=final_winner,
            win_prize_part=win_prize_part,
            internal_jackpot_part=internal_jackpot_part,
            additional_jackpot_part=additional_jackpot_part,
            rectification_progressive_number=rectification_progressive_number,
            total_win_prize=total_win_prize,
            total_internal_jackpot=total_internal_jackpot,
            total_additional_jackpot=total_additional_jackpot,
            win_list=win_list,
        )

    def validate_and_prepare_data_260(
        self,
        play_code: int,
        session_id: str,
        total_winner_number: int,
        initial_winner: int,
        final_winner: int,
        win_prize_part: float,
        internal_jackpot_part: float,
        additional_jackpot_part: float,
        rectification_progressive_number: int,
        total_win_prize: float,
        total_internal_jackpot: float,
        total_additional_jackpot: float,
        win_list: List[WinObject],
    ):
        self.check_play_code(play_code)
        self.check_aams_id(session_id)
        self.check_type("total_winner_number", total_winner_number, int)
        self.check_type("initial_winner", initial_winner, int)
        self.check_type("final_winner", final_winner, int)
        self.check_positive_amount("win_prize_part", win_prize_part)
        self.check_positive_amount("internal_jackpot_part", internal_jackpot_part)
        self.check_positive_amount("additional_jackpot_part", additional_jackpot_part)
        self.check_type("rectification_progressive_number", rectification_progressive_number, int)
        self.check_positive_amount("total_win_prize", total_win_prize)
        self.check_positive_amount("total_internal_jackpot", total_internal_jackpot)
        self.check_positive_amount("total_additional_jackpot", total_additional_jackpot)

        # Controlla che la lista vincitori non contenga più di 1000 elementi
        self.check_prizes_list("win_list (vincite)", win_list, 1000)

        self.data = {
            "nmCodiceGioco": play_code,
            "idSessAams": session_id,
            "nmTotVincitori": total_winner_number,
            "nmInizVincita": initial_winner,
            "nmFinVincita": final_winner,
            "vlImpoParzVincite": self.convert_float_value_to_cent(win_prize_part),
            "vlImpoParzJckInterno": self.convert_float_value_to_cent(internal_jackpot_part),
            "vlImpoParzJckAgg": self.convert_float_value_to_cent(additional_jackpot_part),
            "nmProgrRettifica": rectification_progressive_number,
            "vlImpoTotVincite": self.convert_float_value_to_cent(total_win_prize),
            "vlImpoTotJckInterno": self.convert_float_value_to_cent(total_internal_jackpot),
            "vlImpoTotJckAgg": self.convert_float_value_to_cent(total_additional_jackpot),
            "vincite": self.prepare_winner_list(win_list),
        }

    def handle_message_260_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 260.

        Parametri in risposta:
        =====================
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                return {}
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)
            except Exception as e:
                self.raise_response_generic_error(e, msg_or_exception)
        raise msg_or_exception

    def prepare_winner_list(self, win_list: List[WinObject]) -> List:
        # Ordina la lista vincitori in ordine decrescente rispetto all'importo totale della vincita
        win_list.sort(key=lambda x: x.get("total_prize"), reverse=True)
        try:
            result = []
            for win in win_list:
                played_aams_id = win["played_aams_id"]
                total_prize = win["total_prize"]
                internal_jackpot = win.get("internal_jackpot", 0)
                additional_jackpot = win.get("additional_jackpot", 0)

                self.check_aams_id(played_aams_id, "played_aams_id")
                self.check_positive_amount("total_prize", total_prize)
                self.check_positive_amount("internal_jackpot", internal_jackpot)
                self.check_positive_amount("additional_jackpot", additional_jackpot)

                result.append({
                    "idBiglietto": played_aams_id,
                    "vlImpoTotale": self.convert_float_value_to_cent(total_prize),
                    "vlImpoJckInterno": self.convert_float_value_to_cent(internal_jackpot),
                    "vlImpoJckAgg": self.convert_float_value_to_cent(additional_jackpot),
                })
            return result
        except KeyError as e:
            raise ValidationError({str(e): "This key is required"}) from e
        except Exception as e:
            raise ValidationError(str(e)) from e

# =============================================================================================================
    # 280
    def accredito_vincita_280(
        self,
        id_trans: int,
        play_code: int,
        session_id: str,
        gol_player_token: str,
        player_account_code: str,
        credit_value: float,
        credit_date: datetime,
        credit_progressive_number: int,
        played_aams_id: str,
    ):
        """
        Richiesta di pagamento della vincita tramite accredito sul conto di gioco

        :param id_trans: ID univoco della transazione (assegnato dal fornitore)
        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param session_id: (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        :param gol_player_token (stAccesso) codice di accesso (token) assegnato al giocatore sulla piattaforma GOL
        :param player_account_code (stConto) codice conto del giocatore
        :param credit_value (vlImpoVincita) importo accreditato
        :param credit_date (xdAccredito) data accredito vincita (UTC)
        :param credit_progressive_number (nmProgrAccredito) progressivo di accredito. Vale 1 per il primo accredito
        :param played_aams_id (idBiglietto) codice univoco del diritto di partecipazione assegnato dal sistema AAMS

        """
        return self._do_request_message(
            message_code=280,
            id_trans=id_trans,
            play_code=play_code,
            session_id=session_id,
            gol_player_token=gol_player_token,
            player_account_code=player_account_code,
            credit_value=credit_value,
            credit_date=credit_date,
            credit_progressive_number=credit_progressive_number,
            played_aams_id=played_aams_id,
        )

    def validate_and_prepare_data_280(
        self,
        play_code: int,
        session_id: str,
        gol_player_token: str,
        player_account_code: str,
        credit_value: float,
        credit_date: datetime,
        credit_progressive_number: int,
        played_aams_id: str,
    ):
        self.check_play_code(play_code)
        self.check_aams_id(session_id)
        self.check_gol_player_token(gol_player_token)
        self.check_player_account_code(player_account_code)
        self.check_positive_amount("credit_value", credit_value)
        self.check_utc_timezone("credit_date (xdAccredito)", credit_date)
        self.check_type("credit_progressive_number", credit_progressive_number, int)
        self.check_aams_id(played_aams_id, "played_aams_id (idBiglietto)")

        self.data = {
            "nmCodiceGioco": play_code,
            "idSessAams": session_id,
            "stAccesso": gol_player_token,
            "stConto": player_account_code,
            "vlImpoVincita": self.convert_float_value_to_cent(credit_value),
            "xdAccredito": strftime(credit_date),
            "nmProgrAccredito": credit_progressive_number,
            "idBiglietto": played_aams_id,
        }

    def handle_message_280_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 280.

        Parametri in risposta:
        =====================
        account_amount (vlSaldoCarta) saldo conto di gioco
        account_bonus_amount (vlSaldoBonus) (opzionale) saldo real bonus
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            try:
                response_dict = {
                    "account_amount": self.convert_cent_value_to_float(msg_or_exception["vlSaldoCarta"]),
                    "account_bonus_amount": self.convert_cent_value_to_float(msg_or_exception.get("vlSaldoBonus", 0)),
                }
                return response_dict
            except KeyError as e:
                self.raise_response_missing_field_error(str(e), msg_or_exception)
            except Exception as e:
                self.raise_response_generic_error(e, msg_or_exception)
        raise msg_or_exception

# =============================================================================================================
    # 300
    def fine_sessione_300(
        self,
        id_trans: int,
        play_code: int,
        session_id: str,
        close_session_datetime: datetime,
        was_session_valid: bool = False,
    ):
        """
        Conclude la sessione di gioco

        :param id_trans: ID univoco della transazione (assegnato dal fornitore)
        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param session_id: (idSessAams) identificativo univoco della sessione sui sistemi AAMS
        :param close_session_datetime (xdChiusura) timestamp della chiusura sessione (UTC)
        :param was_session_valid (flChiusuraSess) SI se per la sessione era stata inviata la convalida, NO altrimenti

        """
        return self._do_request_message(
            message_code=300,
            id_trans=id_trans,
            play_code=play_code,
            session_id=session_id,
            close_session_datetime=close_session_datetime,
            was_session_valid=was_session_valid,
        )

    def validate_and_prepare_data_300(
        self,
        play_code: int,
        session_id: str,
        close_session_datetime: datetime,
        was_session_valid: bool = False,
    ):
        self.check_play_code(play_code)
        self.check_aams_id(session_id)
        self.check_utc_timezone("close_session_datetime (xdChiusura)", close_session_datetime)
        self.check_type("was_session_valid", was_session_valid, bool)

        self.data = {
            "nmCodiceGioco": play_code,
            "idSessAams": session_id,
            "xdChiusura": strftime(close_session_datetime),
            "flChiusuraSess": self.convert_bool_to_fl(was_session_valid),
        }

    def handle_message_300_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 300.

        Parametri in risposta:
        =====================
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            return {}
        raise msg_or_exception

# =============================================================================================================
    # 830
    def software_installato_830(
        self,
        id_trans: int,
        play_code: int,
        version: str,
        modules: List[str],
    ):
        """
        Comunicazione degli estremi identificativi dei file critici.

        TODO: se gestiti più di un gioco, modificare il metodo.

        :param id_trans: ID univoco della transazione (assegnato dal fornitore)
        :param play_code: (nmCodiceGioco) codice gioco AAMS
        :param version: (stVersione) versione software (nella forma major.minor)
        :param modules: (moduli) lista di oggetti modulo (ciascuno rappresenta un file critico)
        """
        return self._do_request_message(
            message_code=830,
            id_trans=id_trans,
            play_code=play_code,
            version=version,
            modules=modules,
        )

    def validate_and_prepare_data_830(
        self,
        play_code: int,
        version: str,
        modules: List[str],
    ) -> None:
        self.check_play_code(play_code)
        self.check_version(version)
        self.data = {"elementi": [self.prepare_element_object(play_code, version, modules)]}

    def prepare_element_object(
        self,
        play_code: int,
        version: str,
        modules: List[str],
    ):
        return {
            "nmCodiceGioco": play_code,
            "stVersione": version,
            "moduli": [self.prepare_module_object(module) for module in modules]
        }

    @staticmethod
    def prepare_module_object(module: str) -> dict:
        with open(module, "r", encoding="utf-8") as _module:
            return {
                "stPath": module,
                "stHash": sha1(_module.read().encode("utf-8")).hexdigest()
            }

    def handle_message_830_response(self, response: Response) -> dict:
        """Gestisce la risposta alla richiesta con messaggio 830.

        Parametri in risposta:
        =====================
        La risposta non prevede alcun contenuto.
        """
        success, msg_or_exception = self.handle_generic_message_response(response)
        if success:
            return {}
        raise msg_or_exception

# ================================================================================================================

    # RESPONSE HANDLING
    def handle_generic_message_response(self, response: Response) -> tuple:
        """
        Ogni richiesta prevede una risposta.

        La risposta è un oggetto JSON contenente le chiavi status, che ne descrive l'esito, e response, che ne
        costituisce il vero e proprio payload.

        valore di status        significato

        0                       richiesta andata a buon fine
        1                       errore sui sistemi dell'operatore
        2                       errore AAMS
        3                       impossibile verificare esito richiesta
        4                       impossibile inviare richiesta all'operatore
        5                       impossibile reperire risposta (richiesta mai inviata)
        6                       impossibile reperire risposta (richiesta pending)
        7                       impossibile reperire risposta (richiesta completata)
        8                       errore sui sistemi dell'operatore (non decodificabile)

        In caso di errore 1 o 2, response contiene le chiavi nmFaultCode e stFaultDescription, che danno una
        descrizione numerica e testuale dell'errore.

        L'errore 1013 viene fornito quando i sistemi GOL non hanno ricevuto la risposta dai concessionari in
        tempi accettabili. In questo caso il fornitore è tenuto a riproporre la richiesta utilizzando gli stessi
        parametri (in particolare id_trans dovrà essere il medesimo).

        In caso di errore, i possibili raise sono:
        - GiocaOnlineInternalError (errore non gestito (ad esempio campi richiesti mancanti nella risposta di GOL))
        - AAMSException (errore gestito)
        """
        response_data = self.decode_response_body(response)
        status = self._get_generic_message_response_status(response_data)
        return self._handle_response_data(response_data=response_data, status=status)

    def _get_generic_message_response_status(self, response_data) -> Union[int, Any]:
        if not isinstance(response_data, dict):
            exc = GiocaOnlineInternalError(
                f"Expected GiocaOnline response body to be in JSON format "
                f"during processing of message {self.current_message} this result was obtained.\n"
                f"ID TRANSACTION: {self.id_transaction}.\n"
                f"POST DATA: {self.data}.\n"
                f"Complete Response body: {response_data}."
            )
            logger.error(exc)
            raise exc

        status = 0
        try:
            status = response_data.pop("status")
        except KeyError:
            self.raise_response_missing_field_error("status", response_data)

        if not isinstance(status, int):
            try:
                status = int(status)
            except ValueError:
                pass
        return status

    def _handle_response_data(self, response_data, status):
        if status == GiocaOnlineErrorTypes.GOAL_ERROR_1010.value:
            # Tipicamente response=null
            error_description = self._get_gol_error_type_description(status)
            self.log_error_response(status, error_description, response_data)
            return False, self.get_aams_message_error(status, error_description)

        # A volte GOL ritorna stato 0 anche se il risultato è un errore. Questo avviene ad esempio se si tenta di
        # convalidare una sessione già convalidata. Preferiamo gestire autonomamente gli errori e identificarli come
        # tali.
        if "nmFaultCode" in response_data.get("response", {}):
            status = GiocaOnlineErrorTypes.AAMS_ERROR.value

        if status == GiocaOnlineErrorTypes.OK.value:
            self.log_success_response(status, response_data)
            return True, response_data["response"]

        elif status == GiocaOnlineErrorTypes.OPERATOR_ERROR.value or status == GiocaOnlineErrorTypes.AAMS_ERROR.value:
            error_code, error_description = 0, "InternalError"
            try:
                error_response = response_data.pop("response") or {}

                try:
                    error_code = error_response.pop("nmFaultCode")
                except KeyError:
                    self.raise_response_missing_field_error("nmFaultCode", error_response)

                if not isinstance(error_code, int):
                    try:
                        error_code = int(error_code)
                    except ValueError:
                        pass

                try:
                    error_description = error_response.pop("stFaultDescription")
                except KeyError:
                    self.raise_response_missing_field_error("stFaultDescription", error_response)

            except KeyError:
                self.raise_response_missing_field_error("response", response_data)

            self.log_error_response(status, error_description, response_data)

            return False, self.get_aams_message_error(error_code, error_description)
        else:
            if status == GiocaOnlineErrorTypes.GOAL_ERROR_1013.value:
                self._resend_request()
                return self.handle_generic_message_response(self.response)
            else:
                exc = GiocaOnlineInternalError(f"Request returned status value {status} during processing of "
                                               f"message {self.current_message}.\n"
                                               f"ID TRANSACTION: {self.id_transaction}.\n"
                                               f"ERROR STATUS: {status}\n"
                                               f"ERROR DESCRIPTION: {self._get_gol_error_type_description(status)}\n"
                                               f"POST DATA: {self.data}.\n"
                                               f"Complete Response body: {response_data}")
                logger.error(exc)
                raise exc

    def log_success_response(self, status, response_data):
        logger.info(
            "GOL REQUEST SUCCESS.\n"
            "ID TRANSACTION: %s\n"
            "ERROR STATUS: %s\n"
            "POST DATA: %s.\n"
            "RESPONSE BODY: %s",
            self.id_transaction,
            status,
            str(self.data),
            str(response_data)
        )

    def log_error_response(self, status, error_description, response_data):
        logger.error(
            "AN AAMS ERROR OCCURRED.\n"
            "ID TRANSACTION: %s\n"
            "ERROR STATUS: %s\n"
            "ERROR DESCRIPTION: %s\n"
            "POST DATA: %s.\n"
            "RESPONSE BODY: %s",
            self.id_transaction,
            status,
            error_description,
            str(self.data),
            str(response_data)
        )

    @staticmethod
    def _get_gol_error_type_description(esito: int):
        try:
            return GiocaOnlineErrorTypes(esito).get_description()
        except ValueError:
            return "Invalid status"

    def _resend_request(self):
        """
        Invia nuovamente una richiesta identica a GiocaOnline.
        Questo metodo deve essere chiamato solo se l'errore ricevuto è il 1013
        """
        time.sleep(self.resend_request_sleep)
        self.current_message = self._current_message
        logger.info("RESEND REQUEST MESSAGE %s", self.current_message)
        self.do_request(self.method, self.uri, **self.kwargs)
        self.rprint(self.response)

    def raise_response_missing_field_error(self, key, response_body):
        exc = GiocaOnlineInternalError(
            f"Expected GiocaOnline response body to have required key `{key}`.\n"
            f"During processing of message {self.current_message} this key was missing.\n"
            f"ID TRANSACTION: {self.id_transaction}.\n"
            f"POST DATA: {self.data}.\n"
            f"Complete Response body: {response_body}"
        )
        logger.error(exc)
        raise exc

    def raise_response_generic_error(self, exception, response_body):
        exc = GiocaOnlineInternalError(
            f"During processing of message {self.current_message} an error processing response data occurred.\n"
            f"Giocaonline generic error: {str(exception)}\n"
            f"ID TRANSACTION: {self.id_transaction}.\n"
            f"POST DATA: {self.data}.\n"
            f"Complete Response body: {response_body}"
        )
        logger.error(exc)
        logger.error(exception)
        raise exc from exception

    @staticmethod
    def convert_float_perc_to_cent(perc: float) -> int:
        """
        Converte la percentuale di tipo float in percentuale centesimale.
        Ad esempio: 5% --> 0.05 --> 500
        """
        perc = perc * 10000
        return int(round(perc, 0))

    def convert_float_value_to_cent(self, value: float) -> int:
        """
        Converte il valore numerico di tipo float (o intero) in percentuale centesimale.
        Ad esempio: 10 € --> 10.00 € --> 1000 centesimi di €
        """
        self.check_decimal_digits("Any decimal value", value, 2, "le")
        value = value * 100
        return int(round(value, 2))

    @staticmethod
    def convert_cent_value_to_float(value: int) -> float:
        """
        Converte il valore numerico di tipo intero espresso in centesimi in float.
        Ad esempio: 1000 centesimi di € --> 10.00 € --> 10 €
        """
        if not isinstance(value, int):
            try:
                value = int(value)
            except (ValueError, TypeError):
                raise ValueError("Invalid int %s" % value)
        value = value / 100
        return round(value, 2)

    @staticmethod
    def convert_bool_to_fl(value: bool) -> str:
        """
        Converte il valore booleano nelle stringhe "SI" o "NO"
        """
        assert isinstance(value, bool), f"{value} is not of boolean type"
        return "SI" if value else "NO"

    @staticmethod
    def get_enum_value(value: Any):
        return value.value if isinstance(value, Enum) else value

    # COMMON PARAMETERS VALIDATION
    @staticmethod
    def check_type(name: str, value: Any, _type: Union[type, List[type]]):
        if not isinstance(_type, list):
            _type = [_type]

        if not any(isinstance(value, _t) for _t in _type):
            raise AssertionError(f"{name} must be of types: {str(_type)}")

    def check_aams_id(self, session_id: str, name: str = 'session_id (idSessConc)') -> None:
        """
        `session_id` deve essere composta da esattamente 16 caratteri.
        Se nullo deve corrispondere a 16 spazi vuoti
        """
        self.check_type(name, session_id, str)
        self.check_length(name, session_id, 16, mode="e")

    def check_play_code(self, play_code: int):
        self.check_type("play_code", play_code, int)

    def check_gol_player_token(self, gol_player_token: str):
        self.check_type('gol_player_token', gol_player_token, str)

    def check_player_account_code(self, player_account_code: str):
        self.check_type("player_account_code", player_account_code, str)

    def check_utc_timezone(self, name: str, date: datetime):
        self.check_timezone(name, date, pytz.utc)

    def check_positive_amount(self, name: str, amount: float):
        """
        Controlla che l'importo sia maggiore o uguale a zero e che non abbia più di due decimali.
        """
        self.check_type(name, amount, [int, float])
        self.check_numeric_value(name, amount, 0, "ge")
        self.check_decimal_digits(name, amount, 2, "le")

    def check_perc_amount(self, name: str, amount: float):
        """
        Controlla che l'importo sia maggiore o uguale a zero e che non abbia più di quattro decimali.
        """
        self.check_type(name, amount, [int, float])
        self.check_numeric_value(name, amount, 0, "ge")
        self.check_decimal_digits(name, amount, 4, "le")

    @staticmethod
    def check_version(version: str):
        try:
            major, minor = version.split(".")  # noqa
        except (ValueError, AttributeError):
            raise ValidationError("Version value (%s) is not in the valid format: major.minor" % version)

    def check_related_networks(self, related_networks):
        if related_networks is not None:
            self.check_type("related_networks", related_networks, list)
            for n, el in enumerate(related_networks):
                self.check_type("Related network ID in position %s" % n, el, int)
